/*
   tcluser.c -- handles:
     Tcl stubs for the user-record-oriented commands

   dprintf'ized, 1aug96
*/
     
#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include "../lush.h"
#include "eggdrop.h"
#include "users.h"
#include "proto.h"
#include "cmdt.h"
#include "tclegg.h"

/* eggdrop always uses the same interpreter */
extern Tcl_Interp *interp;
extern struct userrec *userlist;
extern int default_flags;
extern struct dcc_t dcc[];
extern int dcc_total;
extern char origbotname[];

/***********************************************************************/

int tcl_countusers STDVAR
{
  char s[40];
  BADARGS(1,1,"");
  sprintf(s,"%d",count_users(userlist));
  Tcl_AppendResult(irp,s,NULL);
  return TCL_OK;
}

int tcl_validuser STDVAR
{
  BADARGS(2,2," handle");
  if (is_user(argv[1])) Tcl_AppendResult(irp,"1",NULL);
  else Tcl_AppendResult(irp,"0",NULL);
  return TCL_OK;
}

int tcl_finduser STDVAR
{
  char s[20];
  BADARGS(2,2," nick!user@host");
  get_handle_by_host(s,argv[1]);
  Tcl_AppendResult(irp,s,NULL);
  return TCL_OK;
}

int tcl_passwdOk STDVAR
{
  BADARGS(3,3," handle passwd");
  if (pass_match_by_handle(argv[2],argv[1])) Tcl_AppendResult(irp,"1",NULL);
  else Tcl_AppendResult(irp,"0",NULL);
  return TCL_OK;
}

int tcl_chattr STDVAR
{
  int atr,oatr,f,i,pos=1,recheck=0; char s[20];
  BADARGS(2,3," handle ?changes?");
  if ((argv[1][0]=='*') || (!is_user(argv[1]))) {
    Tcl_AppendResult(irp,"*",NULL);
    return TCL_OK;
  }
  oatr=atr=get_attr_handle(argv[1]);
  if (argc==3) {
    /* make changes */
    for (i=0; i<strlen(argv[2]); i++) {
      if (argv[2][i]=='+') pos=1;
      else if (argv[2][i]=='-') pos=0;
      else {
	s[1]=0; s[0]=argv[2][i]; f=str2flags(s)&(~USER_BOT);
	atr=pos?(atr|f):(atr&~f);
	if (f & (USER_OP|USER_DEOP|USER_KICK)) recheck=1;
      }
    }
    atr=sanity_check(atr);
    set_attr_handle(argv[1],atr);
    atr=check_dcc_attrs(argv[1],atr,oatr);
  }
  /* retrieve current flags and return them */
  flags2str(atr,s);
  Tcl_AppendResult(irp,s,NULL);
  if (recheck) recheck_channels();
  return TCL_OK;
}

int tcl_matchattr STDVAR
{
  int i,f,ok=1; char s[2];
  BADARGS(3,3," handle flags");
  for (i=0; i<strlen(argv[2]); i++) {
    s[1]=0; s[0]=argv[2][i]; f=str2flags(s);
    if ((!f) || ((get_attr_handle(argv[1])&f) != f)) ok=0;
  }
  if (ok) Tcl_AppendResult(irp,"1",NULL);
  else Tcl_AppendResult(irp,"0",NULL);
  return TCL_OK;
}

int tcl_adduser STDVAR
{
  BADARGS(3,3," handle hostmask");
  if (is_user(argv[1])) {
    Tcl_AppendResult(irp,"0",NULL);
    return TCL_OK;
  }
  userlist=adduser(userlist,argv[1],argv[2],"-",default_flags);
  Tcl_AppendResult(irp,"1",NULL);
  return TCL_OK;
}

int tcl_addbot STDVAR
{
  BADARGS(3,3," handle address");
  if (is_user(argv[1])) {
    Tcl_AppendResult(irp,"0",NULL);
    return TCL_OK;
  }
  userlist=adduser(userlist,argv[1],"none","-",USER_BOT);
  set_handle_info(userlist,argv[1],argv[2]);
  Tcl_AppendResult(irp,"1",NULL);
  return TCL_OK;
}

int tcl_deluser STDVAR
{
  char s[10];
  BADARGS(2,2," handle");
  sprintf(s,"%d",deluser(argv[1]));
  Tcl_AppendResult(irp,s,NULL);
  return TCL_OK;
}

int tcl_addhost STDVAR
{
  BADARGS(3,3," handle hostmask");
  addhost_by_handle(argv[1],argv[2]);
  return TCL_OK;
}

int tcl_delhost STDVAR
{
  BADARGS(3,3," handle hostmask");
  if (delhost_by_handle(argv[1],argv[2])) {
    Tcl_AppendResult(irp,"1",NULL);
    return TCL_OK;
  }
  Tcl_AppendResult(irp,"0",NULL);
  return TCL_OK;
}

int tcl_getinfo STDVAR
{
  char s[161];
  BADARGS(2,2," handle");
  if (get_attr_handle(argv[1]) & USER_BOT) return TCL_OK;  /* bot */
  get_handle_info(argv[1],s);
  if (s[0]=='@') strcpy(s,&s[1]);
  Tcl_AppendResult(irp,s,NULL);
  return TCL_OK;
}

int tcl_getaddr STDVAR
{
  char s[161];
  BADARGS(2,2," handle");
  if (!(get_attr_handle(argv[1]) & USER_BOT)) return TCL_OK;  /* non-bot */
  get_handle_info(argv[1],s);
  Tcl_AppendResult(irp,s,NULL);
  return TCL_OK;
}

int tcl_getdccdir STDVAR
{
  char s[161];
  BADARGS(2,2," handle");
  get_handle_dccdir(argv[1],s);
  Tcl_AppendResult(irp,s,NULL);
  return TCL_OK;
}

int tcl_getcomment STDVAR
{
  char s[161];
  BADARGS(2,2," handle");
  get_handle_comment(argv[1],s);
  Tcl_AppendResult(irp,s,NULL);
  return TCL_OK;
}

int tcl_getemail STDVAR
{
  char s[161];
  BADARGS(2,2," handle");
  get_handle_email(argv[1],s);
  Tcl_AppendResult(irp,s,NULL);
  return TCL_OK;
}

int tcl_getxtra STDVAR
{
  BADARGS(2,2," handle");
  Tcl_AppendResult(irp,get_handle_xtra(argv[1]),NULL);
  return TCL_OK;
}

int tcl_setinfo STDVAR
{
  BADARGS(3,3," handle info");
  set_handle_info(userlist,argv[1],argv[2]);
  return TCL_OK;
}

int tcl_setdccdir STDVAR
{
  BADARGS(3,3," handle dccdir");
  set_handle_dccdir(userlist,argv[1],argv[2]);
  return TCL_OK;
}

int tcl_setcomment STDVAR
{
  BADARGS(3,3," handle comment");
  set_handle_comment(userlist,argv[1],argv[2]);
  return TCL_OK;
}

int tcl_setemail STDVAR
{
  BADARGS(3,3," handle email");
  set_handle_email(userlist,argv[1],argv[2]);
  return TCL_OK;
}

int tcl_setxtra STDVAR
{
  BADARGS(3,3," handle xtra");
  set_handle_xtra(userlist,argv[1],argv[2]);
  return TCL_OK;
}

int tcl_getlaston STDVAR
{
  char s[21]; time_t t;
  BADARGS(2,2," handle");
  get_handle_laston(argv[1],&t);
  sprintf(s,"%lu",t);
  Tcl_AppendResult(irp,s,NULL);
  return TCL_OK;
}

int tcl_setlaston STDVAR
{
  BADARGS(2,2," handle");
  set_handle_laston(argv[1],time(NULL));
  return TCL_OK;
}

int tcl_userlist STDVAR
{
  struct userrec *u=userlist; int f=0;
  BADARGS(1,2," ?flags?");
  if (argc==2) {
    f=str2flags(argv[1]);
    if ((f==0) && (argv[1][0]) && (argv[1][0]!='-')) return TCL_OK;
  }
  while (u!=NULL) {
    if ((u->flags & f)==f) Tcl_AppendElement(interp,u->handle);
    u=u->next;
  }
  return TCL_OK;
}

int tcl_save STDVAR
{
  write_userfile();
  return TCL_OK;
}

int tcl_reload STDVAR
{
  reload();
  return TCL_OK;
}

int tcl_gethosts STDVAR
{
  struct userrec *u; struct eggqueue *q;
  BADARGS(2,2," handle");
  u=get_user_by_handle(userlist,argv[1]);
  if (u==NULL) return TCL_OK;
  q=u->host; while (q!=NULL) {
    Tcl_AppendElement(irp,q->item);
    q=q->next;
  }
  return TCL_OK;
}

int tcl_chpass STDVAR
{
  char par[10],pass[10];
  BADARGS(3,3," handle password");
  strncpy(par,argv[2],9); par[9]=0;
  nsplit(pass,par);
  change_pass_by_handle(argv[1],pass);
  return TCL_OK;
}

int tcl_chhandle STDVAR
{
  char hand[10]; int x;
  BADARGS(3,3," oldhandle newhandle");
  strncpy(hand,argv[2],9); hand[9]=0;
  if (is_user(hand)) x=0;
  else {
    x=change_handle(argv[1],hand);
    if (x) notes_change(-1,argv[1],hand);
  }
  sprintf(hand,"%d",x); Tcl_AppendResult(irp,hand,NULL);
  return TCL_OK;
}

int tcl_getting_users STDVAR
{
  int i;
  BADARGS(1,1,"");
  for (i=0; i<dcc_total; i++) {
    if ((dcc[i].type==DCC_BOT) && (dcc[i].u.bot->status&STAT_GETTING)) {
      Tcl_AppendResult(irp,"1",NULL);
      return TCL_OK;
    }
  }
  Tcl_AppendResult(irp,"0",NULL);
  return TCL_OK;
}

int tcl_addignore STDVAR
{
  time_t t=time(NULL);
  BADARGS(2,2," nick!user@host");
  prog_ignore(argv[1],t,origbotname);
  return TCL_OK;
}

int tcl_addpermignore STDVAR
{
  BADARGS(2,2," nick!user@host");
  prog_ignore(argv[1],0L,origbotname);
  return TCL_OK;
}

int tcl_delignore STDVAR
{
  int x;
  BADARGS(2,2," nick!user@host");
  x=unprog_ignore(argv[1]);
  if (x) Tcl_AppendResult(irp,"1",NULL);
  else Tcl_AppendResult(irp,"0",NULL);
  return TCL_OK;
}

int tcl_isignore STDVAR
{
  int x;
  BADARGS(2,2," nick!user@host");
  x=match_ignore(argv[1]);
  if (x) Tcl_AppendResult(irp,"1",NULL);
  else Tcl_AppendResult(irp,"0",NULL);
  return TCL_OK;
}
